from __future__ import division
import os
import xml.dom.minidom as minidom

import logging
log=logging.getLogger("Types")
from cuisine.desctypes import *
from cuisine.utils import isxml
from cuisine import framepool
from cuisine.xmlnodebase import xmlnodeclass,xmldocfile,collectiveelement
"""
this module is for import *

"""

# more nodes that got put in their own files:
from channel import *

class Clip(xmlnodeclass):
   
    def essencefile(self,setnode=None):
        return self._getsetorcreate1(Essencefile,setnode)
    def creator(self,setnode=None):
        return self._getsetorcreate1(Creator,setnode)
    def marks(self,setnode=None):
        return self._getsetorcreate1(Marks,setnode)

class Mark(xmlnodeclass):
    """this api is pretty broken. anchored marks need rels to form
    proper ranges, etc.  also the 'frame' attribute is poor, since
    there shouldn't be units in the name.  """
    def __init__(self):
        xmlnodeclass.__init__(self)
        crea=minidom.Element("created")
        crea.setAttribute("time",str(Timestamp()))
        crea.ownerDocument=self.ownerDocument
        self.appendChild(crea)
        self.assignuuid()
    def addname(self,nametype,txt):
        """adds another name for this mark. type is probably master or
        abbrev"""
        e=minidom.Element("name")
        e.setAttribute("type",nametype)
        t=minidom.Text()
        t.ownerDocument=self.ownerDocument
        e.appendChild(t)
        t.data=str(txt)
        self.appendChild(e)
    def setposition(self,pos,anchor=None):
        """sets the one position element of the mark, with optional anchor"""
        e=minidom.Element("position")
        e.setAttribute("frame",str(pos))
        if anchor is not None:
            e.setAttribute("anchor",str(anchor))
        e.ownerDocument=self.ownerDocument
        self.appendChild(e)
    def getname(self,nametype="master"):
        for n in self.getChildrenElementsByTagName("name"):
            if n.getAttribute("type")==nametype:
                text=n.firstChild
                return text.data
    def getposition(self):
        "returns time,anchor for this mark (anchor may be None)"
        pos=self.getChildrenElementsByTagName("position")[0]

        if pos.hasAttribute("anchor"):
            anc=pos.getAttribute("anchor")
        else:
            anc=None
        
        return Duration(pos.getAttribute("frame"),"frames"), anc

class Marks(collectiveelement):
    def childtype(self): return Mark

class Essencefile(xmlnodeclass):
    def __init__(self):
        log.debug("Essencefile.__init__ runs")
        xmlnodeclass.__init__(self)
#        self.setAttribute("myattr","myval")
#        self.myattr="myval"
    def filename(self,newval=None):
        """
        this is the filename from the xml. it might be relative to the
        location of the xml file, or it might be absolute. to get a
        filename to read, use the read-only absolutefilename()
        instead.
        
        don't actually open this file. use media.video.getframe()
        to obtain frames. you would use this function to obtain the
        filename to be the source of a brick."""
        ret=self._getorsetstringnode("filename",newval)
# can't say this, since filename might be relative
#        if not os.path.exists(ret):
#            log.warn("desc file contained a essence filename of '%s', which doesn't exist" % ret)
        return ret

            
    def absolutefilename(self):
        
        """read-only. returns the filename, but makes relative paths
        absolute. remember that the filename path is allowed to be
        relative to the descfile's location."""
        
        fn=self.filename()
        # join() does the right thing if fn is absolute
        descfiledir,descfilename=os.path.split(self.parentbytype(Clip).descfile.docfilename)
        fn=os.path.join(descfiledir,fn)
        return os.path.abspath(fn)

    def creator(self,setnode=None):
        return self._getsetorcreate1(Creator,setnode)
    def media(self,setnode=None):
        return self._getsetorcreate1(Media,setnode)
    def channels(self,setnode=None):
        return self._getsetorcreate1(Channels,setnode)
    

class Filename(xmlnodeclass):
    pass

class Creator(xmlnodeclass):
    def environment(self,setnode=None):
        return self._getsetorcreate1(Environment,setnode)
    def commandline(self,newval=None):
        return self._getorsetstringnode("commandline",newval)
    def software(self,newval=None):
        return self._getorsetstringnode("software",newval)


class Environment(xmlnodeclass):
    def user(self,newuser=None): return self._getorsetattr("user",newuser)
    def host(self,newhost=None): return self._getorsetattr("host",newhost)
    def time(self,newtime=None): return self._getorsetattr("time",newtime)
    
class Framesize(xmlnodeclass):
    def widthheight(self,newwidth=None,newheight=None):
        return (self._getorsettypednode("width",Length,newwidth),
                self._getorsettypednode("height",Length,newheight))
    def aspect(self,newval=None):
        return self._getorsetstringnode("aspect",newval)
    
class Video(xmlnodeclass):
    def framesize(self,setnode=None):
        return self._getsetorcreate1(Framesize,setnode)

    ## why is this in video, rather than in Media??
    def duration(self,newval=None):
        return self._getorsettypednode("duration",Duration,newval)

    def tbd_getframe(self,framenum):
        """temporary api to get the Frame object for a frame number in
        this file (from a cache).

        really, duration sould be moved out of here, and Video and Audio objects
        fetch their respective data (from separate caches, hopefully).
        """
        if not hasattr(self,"_filename"):
            self._filename=self.parentbytype("essencefile").absolutefilename()
        return framepool.fetchframe(self._filename,framenum)

class Audio(xmlnodeclass):
    def tbd_getframe(self,framenum):
        """see Video.tbd_getframe for more notes"""
        if not hasattr(self,"_filename"):
            self._filename=self.parentbytype("essencefile").absolutefilename()
        dvframe=framepool.fetchframe(self._filename,framenum)

        if dvframe is None:
            return None
        from cuisine.player import Player
        samples=Player().audiofromframe(dvframe)
        return samples
        
class Media(xmlnodeclass):
    def video(self,setnode=None):
        return self._getsetorcreate1(Video,setnode)
    def audio(self,setnode=None):
        return self._getsetorcreate1(Audio,setnode)
    def tbd_getframe(self,framenum,component):
        if component=='video':
            return self.video().tbd_getframe(framenum)
        if component=='audio':
            return self.audio().tbd_getframe(framenum)
        raise ValueError("'%s' is not a valid component" % component)

################################################################################

class DescFile(xmldocfile):
    """a description file on disk (not a database version of the same thing).
    this object contains a document, from which you can get the Clip object."""
    def __init__(self,descfilename):
        self._openornew(descfilename,topleveltype=Clip)

    def clip(self):
        "get the Clip that's the top node of the file"
        
        ret=self._gettoplevel()
        if ret is None:
            log.warn("we got a DescFile object and opened the file, but _gettoplevel returned None")
        ret.descfile=self # Clip needs to access the path of the xml file. it shouldn't change.
        return ret


class DescFileManager:
    """this is the class used to manage the desc files themselves.
    it's cheap to open a file many times-- the manager keeps a pool
    of recently-opened files. there's only one such pool, it's shared
    by all DescFileManagers"""

    openedfiles={}# descfilename : DescFile
    
    def __init__(self):
        pass
        
    def newfileforclip(self,clipfilename):
        """creates a new descfile, populated with just the name of the
        clipfile it describes. call save() on the new file right away
        to disk."""
        descfn=clipfilename+".desc"
        if os.path.exists(descfn):
            # maybe we have strategies for alternate names
            # (but we don't, yet)
            raise NameError("desc file already exists")
        df=DescFile(descfn)
        df.clip().essencefile().filename(os.path.abspath(clipfilename))
        self.openedfiles[descfn]=df
        
        return self.openedfiles[descfn]
    def openfilebydescfilename(self,descfilename):
        
        """if you have the desc file name, we can open the clip that
        way. if the file doesn't exist or doesn't parse, exceptions
        will be raised."""

        if descfilename in self.openedfiles:
            return self.openedfiles[descfilename]

        if descfilename.endswith(".desc.desc"):
            descfilename=descfilename[:-5]
            log.warn("auto-fixing filename %s.desc to %s" % ((descfilename,)*2))
        
        if not os.path.exists(descfilename):
            raise IOError("desc file '%s' not found" % descfilename)

        if not isxml(descfilename):
            raise IOError("desc file '%s' is not xml" % descfilename)

        df=DescFile(descfilename)

        # pull the clip filename out of the desc file
        clipfilename=df.clip().essencefile().filename()
        log.debug("the given desc file '%s' is for clip at '%s'" % (descfilename,clipfilename))

        self.openedfiles[descfilename]=df
        return df
        
    def openfileforclip(self,clipfilename):
        """
        this is now the DISFAVORED method. the future of cuisine is
        to always use the desc files (clips might not even be available).

        returns the a DescFile for the given clip disk filename (or
        None if we couldn't get the clipfile). this does not create a
        new file, since this method will (someday) search for missing
        desc files. use newfileforclip to make a new file."""

        descfn=clipfilename+".desc"
        return self.openfilebydescfilename(descfn)

