"""
interim, non object oriented module for doing operations on dv frames. no picture decoding in here, just metadata.
"""

from cuisine.desctypes import Duration, Length
from cuisine import libdv

_dv=libdv.Decoder(0,1,1) # for this module to use to extract headers, etc


def gettimecode(frame):
    """returns a Duration object after asking libdv to extract the
    timecode from the frame. returns None if timecode is not available."""
    _dv.parse_packs(frame)
    
    hh,mm,ss,fr = _dv.get_timestamp_int()
    frames=((hh*60+mm)*60+ss)*30+fr
    
    tc=Duration(int(frames),"frames").as('timecode')
    return tc
    
def getwidthheight(frame):
    """returns w,h Lengths for the frame"""
    _dv.parseheader(frame)
    
    return Length(_dv.width,"px"),Length(_dv.height,"px")
    
def isnewrecording(frame):
    _dv.parseheader(frame)
    _dv.parse_packs(frame)

    return _dv.isnewrecording(frame)

def getrecorddate(frame):

    """ returns an iso8601 string or a 'unknown date' message. the
    length is probably 24 chars, though different timezone formats may
    be different lengths.

    the timezone is coming entirely from strftime, so it might always
    be the local timezone regardless of what's on the tape. this means
    times could be wrong if they're gotten on a computer in a
    different zone from where the tape was recorded- i don't know.
    """
    
    _dv.parseheader(frame)
    _dv.parse_packs(frame)
    try:
        return _dv.getrecorddate() 
    except RuntimeError:
        # libdv failed to find the date
        return "--unknown record date---"
