cimport pyxbuf
import time
from cuisine import pyxbuf

cdef extern from "Python.h":
    object PyBuffer_New(int size)
    int PyObject_AsWriteBuffer(object obj, void **buffer,
                               int *buffer_len) except -1
    object PyString_FromStringAndSize(char *v, int len)
    char* PyString_AsString(object string)
    

cdef extern from "stdlib.h":
    void *malloc(int size)
    void free(void *ptr)

cdef extern from "time.h":
    cdef struct tm:
        int     tm_sec
        int     tm_min
        int     tm_hour
        int     tm_mday
        int     tm_mon
        int     tm_year
        int     tm_wday
        int     tm_yday
        int     tm_isdst

cdef extern from "libdv/dv.h":
    ctypedef enum dv_color_space_t:
        e_dv_color_yuv, 
        e_dv_color_rgb, 
        e_dv_color_bgr0
    
    ctypedef struct dv_decoder_t:
        int quality
        int width
        int height

    ctypedef struct dv_encoder_t:
        int    isPAL
        int    is16x9
        int    vlc_encode_passes
        int    static_qno
        int    force_dct
        int    frame_count
        int    samples_this_frame


    ctypedef signed short int int16_t
    ctypedef unsigned char uint8_t

    dv_decoder_t *dv_decoder_new(int ignored, int clamp_luma,
                                 int clamp_chroma)
    void dv_decoder_free(dv_decoder_t*)

    ctypedef struct FILE
    FILE *dv_set_error_log(dv_decoder_t *dv, FILE *errfile)

    void dv_decode_full_frame(dv_decoder_t *dv, unsigned char *buffer,
                              dv_color_space_t color_space, unsigned
                              char **pixels, int *pitches)
    
    int dv_decode_full_audio(dv_decoder_t *dv, 
                             unsigned char *buffer, int16_t **outbufs)

    int dv_parse_header(dv_decoder_t *dv, unsigned char *buffer)
    void dv_parse_packs(dv_decoder_t *dv, unsigned char *buffer)
    int dv_get_timestamp_int (dv_decoder_t *dv, int *timestamp)
    int dv_get_num_channels(dv_decoder_t *dv)
    int dv_get_frequency (dv_decoder_t *dv)
    int dv_get_num_samples (dv_decoder_t *dv)
    int dv_is_4ch (dv_decoder_t *dv)
    int dv_is_new_recording (dv_decoder_t *dv, uint8_t *buffer)
    int dv_get_recording_datetime_tm (dv_decoder_t *dv,
                                      tm *rec_dt) except 0


    dv_encoder_t *dv_encoder_new     (int rem_ntsc_setup, int clamp_luma,
                                      int clamp_chroma)
    void         dv_encoder_free     (dv_encoder_t* dv_enc)
    int          dv_encode_full_frame(dv_encoder_t *dv_enc, uint8_t **in_,
                                      dv_color_space_t color_space,
                                      uint8_t *out) except -1
    int          dv_encode_full_audio(dv_encoder_t *dv_enc, int16_t **pcm, 
                                      int channels, int frequency,
                                      uint8_t *frame) except -1

    int c_DV_AUDIO_MAX_SAMPLES "DV_AUDIO_MAX_SAMPLES"
    
    int DV_DCT_AUTO
    int DV_DCT_248
    int DV_DCT_88

    int TRUE
    int FALSE

_dct_names = {'auto' : DV_DCT_AUTO,
              '248' : DV_DCT_248,
              '88' : DV_DCT_88}

MAX_CHANNELS = 4
BYTESPERSAMPLE = 2

cdef extern object c_makeaudiobuffers "makeaudiobuffers" (
    int channels, int maxsamplesperchannel)

cdef extern object c_decodeaudio "decodeaudio" (dv_decoder_t *dv,
                                                unsigned char *buffer, 
                                                object audiobufs )
      
DV_AUDIO_MAX_SAMPLES = c_DV_AUDIO_MAX_SAMPLES

cdef dv_color_space_t colorspace_t(s):
    if s == "yuv":
        return e_dv_color_yuv
    if s == "rgb":
        return e_dv_color_rgb
    if s == "bgr0":
        return e_dv_color_bgr0
    
cdef class Decoder:
    cdef dv_decoder_t *dv
    def __new__(self, ignored, clamp_luma, clamp_chroma):
        self.dv = dv_decoder_new(ignored, clamp_luma, clamp_chroma)
        if self.dv == NULL:
            raise ValueError("failed to create dv decoder")

    def __dealloc__(self):
        dv_decoder_free(self.dv)

    property quality:
        def __get__(self):
            return self.dv.quality
        def __set__(self,v):
            self.dv.quality = v

    property width:
        def __get__(self):
            return self.dv.width
    property height:
        def __get__(self):
            return self.dv.height

    def disable_error_log(self):
        dv_set_error_log(self.dv, NULL)

    def decodeframe(self, colorspace, buffer, pyxbuf.Buf outrgb):
        cdef int pitches[3]
        cdef unsigned char *pixels[3]
        cdef char *c_buffer

        if colorspace == "yuv":
            pitches[0] = 720 * 2 # one packed plane, Y0+U0+Y1+V0
            pixels[0] = outrgb.buf
            pitches[1] = pitches[2] = 0
            pixels[1] = pixels[2] = NULL
        elif colorspace == "rgb":
            # packed (one plane) RGB
            pitches[0] = 720 * 3
            pixels[0] = outrgb.buf
            pitches[1] = pitches[2] = 0
            pixels[1] = pixels[2] = NULL
        elif colorspace == "bgr0":
            raise NotImplementedError
        else:
            raise ValueError("unknown colorspace")

        c_buffer = buffer

        #  see http://lists.copyleft.no/pipermail/pyrex/2004-June/000859.html
        #Py_BEGIN_ALLOW_THREADS;
        dv_decode_full_frame(self.dv,
                             <unsigned char *>c_buffer,
                             colorspace_t(colorspace),
                             pixels,
                             pitches)
        #Py_END_ALLOW_THREADS;

    def parseheader(self, frame):
        cdef char *c_buffer
        c_buffer = frame
        return dv_parse_header(self.dv, <unsigned char *>c_buffer)

    def parse_packs(self, frame):
        cdef char *c_buffer
        c_buffer = frame
        dv_parse_packs(self.dv, <unsigned char *>c_buffer)

    def get_num_channels(self):
        return dv_get_num_channels(self.dv)

    def get_frequency(self):
        return dv_get_frequency(self.dv)

    def get_num_samples(self):
        return dv_get_num_samples(self.dv)

    def decodeaudio(self, buffer, audiobufs):
        cdef char *c_buffer
        cdef int16_t *outbufs[4] # MAX_CHANNELS]
        cdef void *vptr
        cdef int buffer_len
        c_buffer = buffer
        
        for i in xrange(len(audiobufs)):
            PyObject_AsWriteBuffer(audiobufs[i], &vptr, &buffer_len)
            if buffer_len < self.get_num_samples() * BYTESPERSAMPLE:
                raise ValueError("buffer length %d is not long enough "
                                 "for %d samples" %
                                 (len, self.get_num_samples()))
            
            outbufs[i] = <int16_t *>vptr

        ret = dv_decode_full_audio(self.dv, <unsigned char *>c_buffer, outbufs)
        if ret != TRUE:
            raise RuntimeError("dv_decode_full_audio failed (%d)" % ret)
        return self.get_num_samples()


    def getrecorddate(self):
        cdef tm t
        dv_get_recording_datetime_tm(self.dv, &t)

        # porly-tested
        t.tm_year = t.tm_year + 1900
        t.tm_mon = t.tm_mon + 1
        
        return time.strftime("%Y-%m-%dT%H:%M:%S%z", 
                             (t.tm_year, t.tm_mon, t.tm_mday,
                              t.tm_hour, t.tm_min, t.tm_sec,
                              t.tm_wday, t.tm_yday, t.tm_isdst))
    def get_timestamp_int(self):
        cdef int ret[4]
        dv_get_timestamp_int(self.dv, ret)
        return ret[0], ret[1], ret[2], ret[3]

    def isnewrecording(self, frame):
        cdef int buffer_len
        c_buffer = frame
        return dv_is_new_recording(self.dv, <unsigned char *>c_buffer)
        
cdef class Encoder:
    cdef dv_encoder_t *dv
    def __new__(self, rem_ntsc_setup, clamp_luma, clamp_chroma):
        self.dv = dv_encoder_new(rem_ntsc_setup, clamp_luma, clamp_chroma)
        if self.dv == NULL:
            raise RuntimeError("failed to create decoder")

    def __dealloc__(self):
        dv_encoder_free(self.dv)

    property isPAL:
        def __get__(self):
            return self.dv.isPAL
        def __set__(self,v):
            self.dv.isPAL = v

    property vlc_encode_passes:
        def __get__(self):
            return self.dv.vlc_encode_passes
        def __set__(self,v):
            self.dv.vlc_encode_passes = v

    property static_qno:
        def __get__(self):
            return self.dv.static_qno
        def __set__(self,v):
            self.dv.static_qno = v

    property force_dct:
        def __get__(self):
            raise NotImplementedError
        def __set__(self,v):
            self.dv.force_dct = _dct_names[v]

    property is16x9:
        def __get__(self):
            return self.dv.is16x9
        def __set__(self,v):
            self.dv.is16x9 = v

    def encodeframe(self, colorspace, pyxbuf.Buf data):
        """returns a pyxbuf.Buf of the frame"""       
        cdef unsigned char *pixels[3]
        cdef pyxbuf.Buf outbuffer
        cdef unsigned char *b

        pixels[0] = data.buf
        pixels[1] = pixels[2] = NULL

        outbuffer = pyxbuf.Buf(120000)
        b = outbuffer.buf
        
        dv_encode_full_frame(self.dv, pixels, colorspace_t(colorspace), b)

        return outbuffer

    def encodeaudio(self, pyxbuf.Buf frame, audio, frequency):
        cdef int16_t *pcm[2]
        cdef int16_t *chanbuf
        cdef int16_t *audio_cast
        cdef int chan
        cdef int samp

        audio_cast = <int16_t *>PyString_AsString(audio)
        
        for chan from 0 <= chan < 2:
            pcm[chan] = <int16_t *>malloc(DV_AUDIO_MAX_SAMPLES*BYTESPERSAMPLE)
            chanbuf = pcm[chan]
            for samp from 0 <= samp < len(audio) / 4:
                chanbuf[samp] = audio_cast[samp * 2 + chan]
                
        try:
            dv_encode_full_audio(self.dv, pcm, 2, frequency, frame.buf)
        finally:
            for chan from 0 <= chan < 2:
                free(pcm[chan])
 
def decodedframesize(colorspace):
    if colorspace == "yuv":
        return 720*480*2
    if colorspace == "rgb":
        return 720*480*3
    if colorspace == "bgr0":
        return 720*480*4

    raise ValueError("unknown colorspace")

def makeaudiobuffers(channels, maxsamplesperchannel):
    bufs = []

    # dv with 12-bit audio would report 2 channels, but then
    # dv_decode_full_audio will try to write to channel=2 anyway.
    channels = 4
    
    for i in range(channels):
        bufs.append(PyBuffer_New(maxsamplesperchannel * BYTESPERSAMPLE))
    return bufs

