
# port of swigbuf.i

cdef extern from "stdlib.h":
    void *malloc(int size)
    void free(void *ptr)

cdef extern from "string.h":    
    void *memcpy(void *dest, void *src, int n)

cdef extern from "Python.h":
    char* PyString_AsString(object string)
    object PyString_FromStringAndSize(char *v, int len)

cdef class Buf:
    def __new__(self, bytes):
        self.buf = <unsigned char*>malloc(bytes)
        if self.buf == NULL:
             raise MemoryError()
        self.length = bytes

    def __str__(self):
        return self.str_length(self.length)

    def __len__(self):
        return self.length

    def str_length(self, length):
        assert length <= self.length
        return PyString_FromStringAndSize( <char *>self.buf, self.length)

    def memcpy(self, inputstring, length=None):
        if length is None:
            length = len(inputstring)
        else:
            assert length <= self.length
            
        memcpy(self.buf, <void *>PyString_AsString(inputstring), length)

    def __dealloc__(self):
        free(self.buf)

def getbuffer(bytes):
    """allocates a plain buffer (unsigned char)"""
    return Buf(bytes)

def buftostring(Buf b, length):
    return b.str_length(length)

def stringintobuffer(Buf b, inputstring, length):
    # read your python string into the already-allocated buffer.
    b.memcpy(inputstring, length)

def freebuffer(Buf b):
    # now using garbage collection
    pass


  
