from gnosis.xml.objectify import XML_Objectify

def readkinosmil(filename):
    """Returns a list of satine xlist nodes for the video clips in the given Kino smil file
    
Kino currently outputs SMIL that looks like this:
<?xml version="1.0"?>
<smil xmlns:smil2="http://www.w3.org/2001/SMIL20/Language">
  <seq>
    <video src="file1.dv" clipBegin="498" clipEnd="540"/>
  </seq>
  <seq>
    <video src="file2.dv" clipBegin="292" clipEnd="400"/>
  </seq>
  ...
</smil>

The returned list will be the <video> nodes only, and by satine's
xlist system each list element has attributes 'src', 'clipBegin', and
'clipEnd' (at least). For example:

  for video in readkinosmil('mysmil.smil'):
     print 'source clip:',video.src
     print 'duration:',(int(video.clipEnd)-int(video.clipBegin)),'frames'

(The printed durations might not be exact, since I don't know if Kino
includes the last frame in the show)

I can't tell how robust the reader is. Future Kino save files might
not work.
    """
    show=[] # the list of dicts
    smilnode = XML_Objectify(filename).make_instance()
    for seq in smilnode.seq:
        # grab the first <video> tag in the <seq>
        videos=seq.video
        try:
            video=videos[0]
        except IndexError:
            print "a seq has no video tag"
        else:
            show.append(videos[0])
    return show

if __name__=='__main__':
    import sys
#    print readkinosmil(sys.argv[1])

    for video in readkinosmil(sys.argv[1]):
        print 'source clip:',video.src
        print 'duration:',(int(video.clipEnd)-int(video.clipBegin)),'frames'
