
from __future__ import nested_scopes,division
import gtk
import gtk.gdk as gdk
import gnomecanvas as canvas
keys=gtk.keysyms
from cuisine.shownodes import *
from cuisine.utils import *
from math import exp
from time import time
import sys
from skim.zooming import Pair
from cuisine.toss import pygtkutil
from louie import dispatcher
import logging
log=logging.getLogger("timeline")
flog=logging.getLogger("timeline-frequent") # events that happen, say, on every mouse motion

import cuisine.toss.timeline
from cuisine.toss.timeline.canvasobj import Canvasobj
from cuisine.toss.timeline.gtkutils import *

class Channelview(Canvasobj):
    """this is a mixin for Brickview which displays channel data over
    the brick and allows the user to edit it. although a channel could
    be discontinuous, this control only supports continuous channels
    for now.

    this object is the control for one channel. in the future, the brickview
    might have a Channelview for each relevant channel."""
    
    def __init__(self):
        "(we know self.master, self.brick, and self.ids have been set)"
        self.yzoom=.5 # yvalues will be scaled up or down. bottom of the track is always zero.
        
        self.construct()


    def construct(self):
        if not self.brick.channels():
            return

        mas = self.master
        chan = self.brick.channels().getall()[0] # <-- this would be selectable when we support multichans
        self.displayedChannel = chan
        segments = chan.getsegments()

        if 'chan' in self.ids:
            for lineid in self.ids.chan[1:]:
                mas.removecanvasobj(lineid)
        if 'handles' in self.ids:
            for handleid in self.ids.handles:
                mas.removecanvasobj(handleid)

        #
        # self.ids.chan = [Channel,lineid1,lineid2,...],
        #   each line is a segment. they're in the order we get
        #   them from the Channel model, which is sorted by x1
        #
        self.ids.chan=[chan]
        for s in segments:
            l = mas.makeline()
            l.connect('event', self.highlightpropertyfactory('width_pixels', 2))
            l.connect('event', self.mousedragfactory(button=1,
                                                     onpress=self.segmentpress, # splits then edits
                                                     onmotion=self.handlemotion,
                                                     onrelease=self.handleRelease))
            self.ids.chan.append(l)
        #
        # self.ids.handles = [box for the start of the first segment,
        #                box for the point between seg0 and seg1,
        #                ...
        #                box for the end of the last segment]
        #
        self.ids.handles=[]
        for seg1, seg2 in zip([None] + segments, segments + [None]):
            handle=mas.makebox()
            handle.connect('event',self.highlightpropertyfactory('width_pixels',4))
            handle.connect('event',self.mousedragfactory(button=1,
                                                         onpress=self.handlepress, # edits the handle
                                                         onmotion=self.handlemotion,
                                                         onrelease=self.handleRelease,
                                                         data=(seg1,seg2)))
            self.ids.handles.append(handle)

    def segmentpress(self, drag, wid, ev):
        """clicked mid-segment
        this bisects the segment and preps the drag data for
        handlemotion to take care of the rest"""

        if not ev.state & gtk.gdk.CONTROL_MASK:
            print "You clicked a segment, which doesn't do anything. Ctrl-click to add a point."
            return

        xcut = self.master.timeforxcoord(ev.x) - self.brick.origin()
        ycut = self.displayedChannel.getchannelvalue(xcut)
        seg1 = self.displayedChannel.findsegment(xcut)
        if not seg1:
            log.warn("splitting a channel, but there was no segment at t=%s" % xcut)
        seg2 = self.displayedChannel.addsegment()

        seg2.x2, seg2.y2 = seg1.x2, seg1.y2
        seg1.x2 = seg2.x1 = xcut
        seg1.y2 = seg2.y1 = ycut

        self.construct()
        self.update()

        # don't attempt to drag from here-- the canvasobjs have all
        # been redone, so the current 'drag' object will not be
        # associated with the right segment

    def handlepress(self, drag, wid, ev):
        "clicked a handle - this means we're starting to edit one or two segments"

        # store original values in drag.startpoints. we only store one point,
        # since we assume a continuous channel.
        s1,s2=drag.data
        if s1 is not None:
            drag.startpoint=Pair(s1.x2,s1.y2)
        else:
            drag.startpoint=Pair(s2.x1,s2.y1) # need this if we're dragging the left end of the channel

    def handlemotion(self, drag, wid, ev, delta):
        "dragging a handle"
        if 'released' in drag:
            return
        
        mas=self.master

        try:
            seg1, seg2 = drag.data
        except TypeError:
            return

        d = Pair(mas.dtimefordxcoord(delta.x), self.dyvaluefordycoord(delta.y))

        if ev.state & gtk.gdk.SHIFT_MASK:
            # shift allows horiz motion only
            d=d.x,0

        newpos = drag.startpoint + d
        newpos = Pair(newpos.x, max(0,newpos.y)) # clamp at 0

        if self.displayedChannel.findsegment(newpos.x) not in [seg1, seg2]:
            # dragged outside the neighbors. You'd expect the x value
            # to clamp at the neighbor's x value, but if seg2 becomes
            # zero-width, the segments sort funny and get
            # reassigned. I'm not sure what the fix is. The current
            # behavior is pretty good- the point will stop moving if
            # you cross a neighbor's boundary.
            pass
        else:
            if seg1 is not None:
                seg1.x2, seg1.y2 = newpos
            if seg2 is not None:
                seg2.x1, seg2.y1 = newpos
        self.update()
        mas.canvas.update_now()

    def handleRelease(self, drag, wid, ev):
        drag.released = True

    def withdraw(self):
        if hasattr(self.ids,'chan'):
            for o in self.ids.chan[1:]:
                self.master.removecanvasobj(o)
        if hasattr(self.ids,'handles'):
            for o in self.ids.handles:
                self.master.removecanvasobj(o)

    def ycoordforvalue(self,yv):
        "returns the canvas coord for the given (normalize 0..1) value on our track"
        tracktop,trackbottom=self.master.ycoordsfortrack(self.brick.track())
        return trackbottom+(yv*self.yzoom)*(tracktop-trackbottom)
    def yvalueforcoord(self,yc):
        """returns the normalized value for the canvas coord (which is
        hopefully somewhere over our track).
        
        this is the inverse of the preceding method."""
        tracktop,trackbottom=self.master.ycoordsfortrack(self.brick.track())
        return (yc-trackbottom)/(tracktop-trackbottom)/self.yzoom
    def dyvaluefordycoord(self,dyc):
        "like the above, but without the offset"
        tracktop,trackbottom=self.master.ycoordsfortrack(self.brick.track())
        return dyc/(tracktop-trackbottom)/self.yzoom

    def update(self):
        if self.brick.channels():
            mas=self.master

            borg=self.brick.origin()
                
            segments=self.ids.chan[0].getsegments()

            for seg,lineid in zip(segments,self.ids.chan[1:]):
                x1,x2=mas.xcoordfortime(seg.x1+borg,seg.x2+borg)
                y1=self.ycoordforvalue(seg.y1)
                y2=self.ycoordforvalue(seg.y2)
                setproperties(lineid,points=[x1,y1,x2,y2])
            for handle,seg1,seg2 in zip(self.ids.handles,[None]+segments,segments+[None]):
                if seg2:
                    x=mas.xcoordfortime(seg2.x1+borg)
                    y=self.ycoordforvalue(seg2.y1)
                else:
                    x=mas.xcoordfortime(seg1.x2+borg)
                    y=self.ycoordforvalue(seg1.y2)
                setproperties(handle,x1=x-3,y1=y-3,x2=x+3,y2=y+3)
                
                
class Brickview(Canvasobj,Channelview):
    """
    a single brick, which is drawn over some track.  brick knows its
    own timings and track, but it asks the Timelinewidget for the
    canvas coordinates for the times and track.

    """

    def __init__(self,master,brick):
        make_attributes_from_args('master','brick')
        self._construct()
        Channelview.__init__(self)
        self.keyboardcommands()

    def __str__(self):
        return "<brickview for '%s'>" % self.brick

    def withdraw(self):
        "removes all the canvas objects"
        Channelview.withdraw(self)
        for o in self.ids.values():
            self.master.removecanvasobj(o)
        if hasattr(self,'markids'): # for clip bricks with marks
            for o in [x[0] for x in self.markids]+[x[3] for x in self.markids]:
                self.master.removecanvasobj(o)
    def keyboardcommands(self):
        mapping={ keys.question : self.slicebrick,
                  keys.asciitilde : self.killbrick,
                  keys.plus : self.addtoselection,
                  keys.at : self.addaudiocurve,
                  }
        def handle(wid,ev):
            if ev.type==gdk.KEY_PRESS:
                if ev.keyval in mapping:
                    x,y,state=ev.window.get_pointer()
                    mousetime=self.master.timeforxcoord(x)
                    mapping[ev.keyval](ev,mousetime)
                
            return 0
        
        self.ids.body.connect('event',handle)

    def addaudiocurve(self,ev,mousetime):
        self.brick.addaudiocurve()
        # redraw
        mas=self.master
        mas.removebrickview(self.brick)
        mas.addbrickview(self.brick)
        mas.update()
        

    def addtoselection(self,ev,mousetime):
        "add this to the timeline's selction list"
        if self not in self.master.selectedbrickviews:
            self.master.selectedbrickviews.append(self)
        log.info("selection list now contains: %s" % ", ".join(map(str,self.master.selectedbrickviews)))
            
    def killbrick(self,ev,mousetime):
        """kills this brick"""
        b=self.brick
        show=b.parentbytype(Show)
        try:
            self.master.removebrickview(self.brick)
        except:
            sys.excepthook(*sys.exc_info())
            log.error("error trying to remove the brick view - brick is not killed")
        else:
            show.bricks().remove(b)

    def slicebrick(self,ev,mousetime):
        """user has requested that we slice this brick into
        two bricks that play the same
        """
        mas=self.master
        newbrick=self.brick.slice(mousetime-self.brick.origin())
        mas.removebrickview(self.brick)
        mas.addbrickview(self.brick)
        mas.addbrickview(newbrick)
        mas.update()
        
    def _construct(self):
        "this creates all the canvas objecs for a new brick and sets up some signal handlers"

        log.debug("constructing view for %s" % self.brick)
        self.ids=ClassyDict() # most canvas objs will be in this structure (see self.withdraw())
        ids=self.ids
        mas=self.master

        self.hidden=0 # this indicates whether the whole brick is out-of-view (see update())

        ids.leftedge=mas.makeline()
        ids.rightedge=mas.makeline()
#        ids.topedge=mas.makeline()
        ids.bottomedge=mas.makeline()
        ids.body=mas.makefilledrect()

        # label for the intime (in show time)
        ids.intime=mas.maketext(text="",anchor="sw")

        # brick's own str() appears on it
        s=str(self.brick)
        if s.rfind('/')!=-1:
            s=s[s.rfind('/')+1:]
        s=s+" intime=%s" % self.brick.intime()
        ids.sourcename=mas.maketext(s,anchor="nw")
        
        log.debug("text width is %r" % (ids.sourcename.get_property("text_width")))
        #        ids.sourcename.set_property("clip_width",150)
        #        ids.sourcename.set_property("clip",1)

        # text for each mark
        ##DISABLED to turn off all mark texts
        if 1 and self.brick.clip():
            self.markids=[] # list of [canvasid,time,text,lineid,hidden]
            for mark in self.brick.clip().marks().getall():
                txt=mark.getname()
                pos=mark.getposition()[0]

                # this is a terrible optimization - only text that is within
                # the brick at the time of _construction will be instantiated.
                # either we should make more texts when the brick gets stretched,
                # or speed up the canvas so this isn't necessary.
                if not self.brick.intime()<pos<self.brick.outtime():
                    continue

                #if len(txt)>25:
                #    txt=txt[:22]+"..."
                     
                self.markids.append( [mas.maketext(),        # canvasid for text
                                      pos                  , # mark time
                                      txt,        # mark text (not wrapped)
                                      mas.makeline(),        # canvasid for line 
                                      0,         # is this mark hidden (0=no, 1=text hidden, 2=both hid/out of sight)
                                      ])
                self.markids[-1][0].set_property("fill_color","blue")

        self.setuphighlighting()
        self.setupdragging()


    def setuphighlighting(self):
        "arranges some mouse-over highlighting"
        ids=self.ids

        for attr in 'leftedge','rightedge':
            getattr(ids,attr).connect('event',self.highlightpropertyfactory("width_pixels",4))

        # (fill_color_gdk,gdk.color_parse("red") doesnt work right)
        ids.body.connect('event',self.highlightpropertyfactory("fill_color_rgba",0xddddddffL))

    def setupdragging(self):
        ids=self.ids

        for canvasobj, brickparam in ((ids.leftedge,'intime'),
                                      (ids.rightedge,'outtime')):

            def moreupdate():
                self.update()
                self.master.canvas.update_now()
            handler=self.mousedragtimefactory(button=1,
                                              getval=getattr(self.brick,brickparam), # getattr returns a function for later use
                                              setval=getattr(self.brick,brickparam),
                                              updatefunc=moreupdate)
            canvasobj.connect('event',handler)

        ## this arranges for the user to be able to adjust the origin and track
        ## by dragging the body of the brick.        
        ##(we'd expect shift key to constrain this effect to just one variable - it doesn't yet)
        def onpress(drag,wid,ev):
            drag.startedInBrick = True # as opposed to a click that started in a segment or handle
            drag.originstart=self.brick.origin()
            drag.trackstart=self.brick.track()

            #multi selection hack
            drag.originstarts=[]
            drag.trackstarts=[]
            for bv in self.master.selectedbrickviews:
                drag.originstarts.append(bv.brick.origin())
                drag.trackstarts.append(bv.brick.track())
        def onmotion(drag,wid,ev,delta):
            if 'startedInBrick' not in drag:
                print "dragging in brick, but click started on something else"
                return
            mas=self.master
            # holding shift disables offsetting bricks, so they change track only
            if not ev.state & gtk.gdk.SHIFT_MASK:
                # change brick origin
                offset=+mas.dtimefordxcoord(delta.x)
                
                if self.brick.bricktype()=='edit':
                    log.warn("changing origin of edit bricks currently disabled")
                else:
                    self.brick.origin(drag.originstart+offset)

                # multi selection hack
                for bv,ostart in zip(self.master.selectedbrickviews,drag.originstarts):
                    bv.brick.origin(ostart+offset)
                
            # change track
            newtrack=mas.trackforycoord(ev.y)
            if newtrack!=self.brick.track():
                self.brick.track(newtrack)
            dispatcher.send("timeline changed") # someday this might be more specific
            self.update()
            for bv in self.master.selectedbrickviews:
                bv.update()
            mas.canvas.update_now()

        ids.body.connect('event',self.mousedragfactory(button=1,
                                                       onpress=onpress,
                                                       onmotion=onmotion))
        

    def update(self,fast=0):
        """puts the brick object in the right spot on the canvas, by
        asking the Timelinewidget where the intime and outtime should
        be drawn"""
        ids=self.ids
        xmarg=0; ymarg=3 # margin between clip extents and the box around it

        # find the show times for the in and out, and the track
        b=self.brick
        intime=b.origin()+b.intime()
        outtime=b.origin()+b.outtime()
        track=b.track()

        # get coords for them
        mas=self.master
        x1,x2=mas.xcoordfortime(intime,outtime)

        xmin,xmax=mas.visiblexcoord()

        if x1>xmax or x2<xmin:
            if self.hidden:
                # culled!
                return
            # otherwise, we need to run this function once more to get the brick out of view
            self.hidden=1
            if fast: # sorry, if we're drawing fast, the obj won't get properly hidden on this run
                self.hidden=0
        else:
            self.hidden=0

        y1,y2=mas.ycoordsfortrack(track)

        # move the graphics to the coords
        for idname,lx1,ly1,lx2,ly2 in (('leftedge',  x1+xmarg,y1+ymarg,x1+xmarg,y2-ymarg),
                                       ('rightedge', x2-xmarg,y1+ymarg,x2-xmarg,y2-ymarg),
#                                       ('topedge',   x1+xmarg,y1+ymarg,x2-xmarg,y1+ymarg),
                                       ('bottomedge',x1+xmarg,y2-ymarg,x2-xmarg,y2-ymarg)
                                       ):
            setproperties(getattr(ids,idname),points=[lx1,ly1,lx2,ly2])

        if not fast:
            setproperties(ids.body,x1=x1+8,y1=y1+6,x2=x2-8,y2=y2-6)
            setproperties(ids.sourcename,x=x1+15,y=y1+7)

            # this label has the show coordinate of the brick intime
            txt="show in="+str((b.intime()+b.origin()).as("timecode"))
    #        txt="in=%r+%r=str(%r)=%r" % (b.intime(),b.origin(),b.intime()+b.origin(),str(b.intime()+b.origin()))
            if x2-x1<50: # drop text when the brick is tiny
                txt=""
            setproperties(ids.intime,text=txt,x=x1+15,y=y2-4)

        # arrange the mark texts
        if hasattr(self,'markids') and not fast:
            bin=self.brick.intime()
            bout=self.brick.outtime()
            borg=self.brick.origin()
            xmin,xmax=mas.visiblexcoord()
            shown,total=0,0
            lastxmark=0
            for i in range(len(self.markids)):
                (id,pos,txt,lineid,hidden)=self.markids[i]
                newhidden=2 # 0 complete, 1 line only, 2 hidden
                if bin<pos<bout:
                    # find the position of the next mark (or bout, if that's closer)
                    if i<len(self.markids)-1:
                        nextpos=min(self.markids[i+1][1],bout)
                    else:
                        nextpos=bout

                    # find the canvas coord of the mark itself and the coord for the next thing
                    x1,x2=mas.xcoordfortime(borg+pos,borg+nextpos)

                    if x2>xmin and x1<xmax: # if any part is visible...
                        if lineid:
                            if x1<lastxmark+3: # line is too close to previous line
                                newhidden=2
                            else:
                                setproperties(lineid,points=[x1,y2-20,x1,y2])
                                newhidden=1
                            lastxmark=x1
                        if x2-x1>30: # enough room for some text too?
                            txt=wrap(txt,(x2-x1)/6) # 6 is a made-up character width
                            setproperties(id,text=txt,x=x1,y=y1+15)
                            newhidden=0

                if hidden<newhidden:
                    if newhidden>=1:
                        setproperties(id,text="")
                    if lineid and newhidden==2:
                        setproperties(lineid,points=[0,0,0,0])
                self.markids[i][4]=newhidden
        Channelview.update(self)

try:
    import psyco
    psyco.bind(Brickview._construct)
    psyco.bind(Brickview.update)
except ImportError:
    pass
