from __future__ import nested_scopes,division
import gtk
import gtk.gdk as gdk
import gnomecanvas as canvas
keys=gtk.keysyms
from cuisine.shownodes import *
from cuisine.utils import *
from math import exp
from time import time
import sys
from skim.zooming import Pair
from cuisine.toss import pygtkutil
from louie import dispatcher
import logging
log=logging.getLogger("timeline")
flog=logging.getLogger("timeline-frequent") # events that happen, say, on every mouse motion
from cuisine.toss.timeline import basefont

class Canvasobj:
    """my highlevel canvas objects, such as brickviews, trackviews, gridlines, cursors.
    this is an ABC.
    """
    def __init__(self,master,*args):
        """all the objs take a master which points back to the Timelinewidget"""
        raise NotImplementedError
    def update(self):
        """this means 're-calc the locations of your objects, since
        something may have changed about where/how they should be
        drawn'"""
        raise NotImplementedError


    def highlightfactory(self,highlightcmd,unhighlightcmd):        
        """an event handler factory that does mouse-over highlighting.
        call this with functions that should be called when a widget
        is entered and left, and this factory will return you an event
        handler that calls your commands upon ENTER_NOTIFY and
        LEAVE_NOTIFY.  your functions will be called with one
        argument: the widget from the signal which you can use if you
        want.

        if highlighting just means setting a widget property to a new value,
        read on to the highlightpropertyfactory() method which does just that.

        example:
            def brighten_func(wid):
                wid.setbackground(bright)
            def darken_func(wid):
                wid.setbackground(dark)
            mywidget.connect('event',highlightfactory(brighten_func, darken_func))
        """

        def handler(wid,ev):
            if ev.type==gdk.ENTER_NOTIFY:
                highlightcmd(wid)
            elif ev.type==gdk.LEAVE_NOTIFY:
                unhighlightcmd(wid)
            return 0 # proceed to next handler
        return handler

    def highlightpropertyfactory(self,propname,prophighlightvalue):
        def enter(wid):
            # rememeber old setting (on the widget, not on self, since
            # a few widgets might be entered at once)
            wid._savehighlightprop=wid.get_property(propname)
            flog.debug("saved old %s '%r'" % (propname,wid._savehighlightprop))
            # change to the highlighted setting
            wid.set_property(propname,prophighlightvalue)
        def leave(wid):
            if not hasattr(wid,'_savehighlightprop'):
                log.debug("we got a LEAVE_NOTIFY on widget %r, but we hadn't highlighted it" % wid)
                return
            flog.debug("restoring %s '%r'" % (propname,wid._savehighlightprop))
            wid.set_property(propname,wid._savehighlightprop)
            del wid._savehighlightprop
        return self.highlightfactory(enter,leave)
    def mousedragfactory(self,button,
                         onpress=lambda *args: None,
                         onmotion=lambda *args: None,
                         onrelease=lambda *args: None,
                         data=None):
        """makes a handler for mouse button drags.

        button - the button number to be sensitive to (no modifer discrimination, yet)

        onpress - function that will be called with the mouse press
        event. this function does not need to save the cursor
        position; that will be handed to the onmove function.
        function prototype is "def onpress(drag,wid,ev):". See below
        for a discussion of the 'drag' paramter.

        onmotion - function that will be called with the mouse motion
        event *only after a successful onpress and before the
        onrelease*. function prototype is "def onmotion(drag,wid,ev,delta):"
        where delta is a Pair describing the offset of the current
        mouse position (in pixels) from the mousedown position.

        onrelease - function that will be called when the button is
        released.  prototype is "def onrelease(drag,wid,ev):". this
        function doesn't have to clean up the drag structure, since
        the drag object is forgotten after this function anyway.

        temporary variables that last for the duration of the drag
        could be stored in the (mostly) empty transient object passed
        in the 'drag' parameter. (the drag object will be stored in
        self._drag and removed after the drag is done. this should be
        okay, as we expect only one drag at once).

        the data you pass into the factory will be planted into drag.data.
        """
        def handler(wid,ev):

            if ev.type==gdk.BUTTON_PRESS and ev.button==button:
                self._drag=ClassyDict()
                drag=self._drag
                drag.data=data
                drag._startmousepos=Pair(ev.x,ev.y)
                onpress(drag,wid,ev)
            elif ev.type==gdk.MOTION_NOTIFY:
                # we only want to react if there's a drag in progress,
                # which will be signified by the existence of a _drag attribute
                if hasattr(self,'_drag'):
                    drag=self._drag
                    delta=Pair(ev.x,ev.y)-drag._startmousepos
                    onmotion(drag,wid,ev,delta)
            elif ev.type==gdk.BUTTON_RELEASE:
                if hasattr(self,'_drag'):
                    onrelease(self._drag,wid,ev)
                    del self._drag
            return 0 # proceed to next handler
        return handler
    def mousedragtimefactory(self,button,getval,setval,updatefunc):
        """a specialized handler for mouse drags that need to adjust a
        time on the timeline. you give the mouse button, a function to
        get the value, a function to set the value, and a function to
        call each time the attribute changes.  this method returns a
        handler to connect to the 'event'. getval is called with no
        args, setval is called with one.

        the 'timeline changed' signal is fired on each motion event.

        """
        def onpress(drag,wid,ev):
            drag.paramstart=getval()
        def onmotion(drag,wid,ev,delta):
            mas=self.master
            newval=drag.paramstart+mas.dtimefordxcoord(delta.x)
            flog.debug("mouse dragtime %r %r" % (setval,newval))
            setval(newval)
            dispatcher.send("timeline changed") # someday this might be more specific
            updatefunc()
        return self.mousedragfactory(button,onpress,onmotion)
