from __future__ import nested_scopes,division
import gtk
import gtk.gdk as gdk
import gnomecanvas as canvas
keys=gtk.keysyms
from cuisine.shownodes import *
from cuisine.utils import *
from math import exp
from time import time
import sys
from skim.zooming import Pair
from cuisine.toss import pygtkutil
from louie import dispatcher
import logging
log=logging.getLogger("timeline")
flog=logging.getLogger("timeline-frequent") # events that happen, say, on every mouse motion
from cuisine.toss.timeline import basefont

"""
todo:
slicebrick, killbrick, and maybe more belong in the shownodes, not in this UI layer.
the shownodes would notify the UI layer of the extent of the changes that happened.

 
"""

import cuisine.toss.timeline
from cuisine.toss.timeline.canvasobj import Canvasobj
from cuisine.toss.timeline.gtkutils import *



class Trackview(Canvasobj):
    """one track on the canvas. this view does not have any notion of
    the time scale. it just draws the title box on the left and any
    decorations that surround the track view. the brickviews are drawn
    on top of the whatever the trackview draws.

    +-----------------------------------------------+
    | title  |     bricks                           |
    |        |                                      |
    +-----------------------------------------------+

    """
    
    def __init__(self,master,track):
        """this view is forever tied to one Track object"""
        make_attributes_from_args('master','track')
        self._construct()

    def _construct(self):
        """makes all the elements of the track object and stores their canvas ids.
        the objects are not in position yet"""

        self.ids=ClassyDict()
        ids=self.ids
        mas=self.master

        # the box around the whole track
        ids.mainoutline = mas.makebox()
        # the separator between the title part and the time part
        ids.dividerbar = mas.makeline()
        ids.tracknametext = mas.maketext(self.track.name(),anchor="center")
        
        mas=self.master


        
    
    def update(self,fast=0):
        """call this to move the track to new canvas coordinates (some coords
        are retrieved from the master)."""
        ids=self.ids
        mas=self.master

        xleft,xtitle,xright=mas.xcoordsfortrack(self.track)
        ytop,ybottom=mas.ycoordsfortrack(self.track)

        ### this cache seems to provide maybe 1.5% speedup during playback
        if hasattr(self,'cachelastcoords') and self.cachelastcoords==(xleft,xtitle,xright,ytop,ybottom):
            return
        self.cachelastcoords=(xleft,xtitle,xright,ytop,ybottom)
        ###
        
        setproperties(ids.mainoutline,
                      x1=xleft,y1=ytop,x2=xright,y2=ybottom)
        setproperties(ids.dividerbar,
                      points=[xtitle,ytop,xtitle,ybottom])
        setproperties(ids.tracknametext,
                      x=(xleft+xtitle)/2,y=(ytop+ybottom)/2)

class Wrappedtext(canvas.CanvasGroup):
    """text subclass that wraps its own lines"""
    def __init__(self,*args,**kwargs):
        #        canvas.CanvasText.__init__(self,*args,**kwargs)
        self.t=self.add(canvas.CanvasText)
        setproperties(self.t,x=kwargs.get('x',0),y=kwargs.get('y',0),
                      font=basefont,fill_color='black',
                      anchor=kwargs.get('gtkanchor',gtk.ANCHOR_NW))
                      
    def set_property(self,prop,val):
        log.debug("setting %r to %r" % (prop,val))
        canvas.CanvasText.set_property(self.t,prop,val)


class Gridlines(Canvasobj):
    """gridlines on seconds"""
    def __init__(self,master):
        self.master=master

        mas=self.master
        self.lines=[]
        self.labels=[]
        for i in range(9):
            t=Duration(i*30,'frames')
            x=mas.xcoordfortime(t)
            self.lines.append(mas.makeline(x,4,x,500))
            self.labels.append(mas.maketext(str(t.as('frames'))))

        [x.lower_to_bottom() for x in self.lines+self.labels]

    def update(self,fast=0):
        mas=self.master
        for i,l,lab in enumerate(self.lines,self.labels):
            x=mas.xcoordfortime(Duration(i*30,'frames'))
            setproperties(l,points=[x,4,x,500])
            setproperties(lab,x=x+2,y=480)

class Timecursor(Canvasobj):
    """a cursor that indicates a position in the show time. user adjustable."""
    time=None
    def __init__(self,master):
        self.master=master
        self.line=self.master.makeline()
        self.lab=self.master.maketext("", font="arial 10")

        self.time=Duration(10,'frames')

        # add highlighting
        self.line.connect('event',self.highlightpropertyfactory("width_pixels",4))

        # add dragability
        self.line.connect('event',self.mousedragtimefactory(button=1,
                                                            getval=lambda: self.time,
                                                            setval=self.changetime,
                                                            updatefunc=self.update))
        self.updatecount=0

    def changetime(self,newtime):
        self.time=newtime
        self.update()
        dispatcher.send("time cursor changed",newtime=newtime)
    time=property(lambda: self.time, changetime)

    def update(self,fast=0):
        mas=self.master
        x=mas.xcoordfortime(self.time)
        setproperties(self.line,points=[x,0,x,750])
        setproperties(self.lab,x=x,y=50,text="%s %s"%(self.time.as('timecode'),self.time.as('frames')))
        for o in self.line,self.lab:
            o.raise_to_top()
        
    def advanceoneframe(self):
        self.skip(Duration(1,'fr'))

    def skip(self,d,update=1):
        self.changetime(self.time+d)


