
from cuisine.utils import make_attributes_from_args
import Tkinter as tk
import cuisine.toss2.ogldraw as draw


from cuisine.toss2.timeline.nodeview import Nodeview

class Trackview(Nodeview):
    """

    one track in the timeline. this is mostly a tk widget on the track
    guide at the left side of the screen, but a trackview also
    involves some opengl material.
    """

    def __init__(self,tkmaster,openglmaster,layout,track):
        """

        this view is tied to the given track object. the tkmaster is
        the left-side tracks pane.  the given Layout object will tell
        us where the track panel is located and how big it is. 
        
        
        """
        Nodeview.__init__(self,tkmaster,openglmaster,layout,track)
        self.track=self.node

        # we introduce ourself to the layout
        layout.newtrackview(self)
        

        y1,y2=layout.windowy_for_track(self.track)
        # we get all rows in range(y1,y2). (not row y2)

        # our frame
        self.trackframe=f=tk.Frame(tkmaster,bd=1,relief='raised')
        # this gets place'd later, in self.redraw

        # the "Track: trackname" display
        w=tk.Frame(f)
        tk.Label(w,text='Track:').pack(side='left')
        e=tk.Entry(w,width=12)
        e.insert(0,self.track.name())
        e.config(state='disabled') # edit this someday? :)
        e.pack(side='left')
        w.pack(side='top',ipadx=10,ipady=10)

        # height adjuster bar
        # eventually, i should be able to pick some tracks and adjust all their heights. that should
        # use the new editing system.
        # also, where's this getting saved?
        heightadjustbarsize=9
        bar=tk.Button(f,relief='groove',bd=heightadjustbarsize//2,height=heightadjustbarsize)
        bar.place(relx=0,relwidth=1,rely=1,height=heightadjustbarsize,anchor='sw')
        def markbar(ev):
            self._marky=ev.y_root
            y1,y2=layout.windowy_for_track(self.track)
            self._markheight=y2-y1
        def adjustbar(ev):
            dy=ev.y_root-self._marky
            layout.changetrackviewheight(self,self._markheight+dy)
        bar.bind("<Button-1>",markbar)
        bar.bind("<B1-Motion>",adjustbar)
       
        self.redraw()
        
    def redraw(self):

        """ if anything's changed about the view, this will be
        called. it updates the coordinates of the trackview.

        this assumes we're set to draw on the opengl surface.
        """
        Nodeview.redraw(self)
        
        # get the top and bottom of our main frame correct
        y1,y2=self.layout.windowy_for_track(self.track)
        self.trackframe.place(relx=0,relwidth=1,y=y1,height=y2-y1-1)

        # the opengl "railings" between tracks
        for y in y1,y2:
            p1=(0,y)
            p2=(self.openglmaster.winfo_width(),y)
            draw.line(p1,p2,(0,0,0))
