"""routines to help with indexing clips- finding them, making thumbnails, etc"""
from __future__ import generators
import os
from cuisine.descnodes import DescFileManager
from cuisine.utils import dirwalk,isxml
import logging
log=logging.getLogger("clipindexing")

_decoder=None # might become a libdv decoder if we need one
def _getdecoder():
    global _decoder
    if not _decoder:
        from cuisine.player import Decoder
        _decoder=Decoder()
    return _decoder

def _clipname_thumbspath(df):
    # returns clipname,thumbspath for the desc file
    clippath,clipname=os.path.split(df.clip().essencefile().absolutefilename())
    return clipname,os.path.join(clippath,"thumbs")

def _thumbimgpath(thumbspath,clipname,framepos):
    framenum=int(framepos.getas("frames"))
    return os.path.join(thumbspath,"%s.%i.jpg" % (clipname,framenum))

def findallclips(basepath="."):
    
    """starting at basepath, this yields descfile objects and their
    paths (relative to basepath) for all desc files found in the tree.
    this is new behavior- we now do not concern ourselves with whether
    the clip essence file is present at all.

    NOTE: this matches any xml file! even ones with desc~ at the end!
    even non-desc files, i think.

    to get the clip path from a resulting descfile, use
    df.clip().essencefile().filename() """

    mgr=DescFileManager()
    for f in dirwalk(basepath,symlinkok=1):
        # try the first line of the file as xml
        if not isxml(file(f)):
            continue

        try:
            df=mgr.openfilebydescfilename(f)
        except ValueError:
            continue

        if df is None:
            continue
        yield df,f
def thumbexists(df,framepos):

    """returns path of thumbnail image or None, if the thumb image is
    not made yet"""

    clipname,thumbspath=_clipname_thumbspath(df)
    
    thumbimgpath=_thumbimgpath(thumbspath,clipname,framepos)
    if os.path.exists(thumbimgpath):
        return thumbimgpath
    else:
        return None
        

def makethumbforframe(df,framepos):
    """
    this decodes the frame (a Duration) and saves a thumbnail in the given path.
    returns the path of the thumb image.
    
    there shouldn't be anything special about thumbs! they should be a conversion
    of the clip, which is usable anywhere the clip is usable. currently, though,
    thumbs are separate jpegs in thumbs/ directories, and they have no desc files.
    """
    clipname,thumbspath=_clipname_thumbspath(df)

    if not os.path.exists(thumbspath):
        os.mkdir(thumbspath)

    thumbimgpath=_thumbimgpath(thumbspath,clipname,framepos)

    if os.path.exists(thumbimgpath):
        return thumbimgpath
    
    framenum=int(framepos.getas("frames"))

    log.info("making thumbnail for %s %s" % (df.clip().essencefile().filename(),framepos))
    frame=df.clip().essencefile().media().video().tbd_getframe(framenum)
    _getdecoder().decode(frame)
    
    cmd="rawtoppm 720 480 - | pnmscale -xsize %i -ysize %i | cjpeg > %s" % (64,48,thumbimgpath)
    (child_stdin, child_stdout)=os.popen2(cmd)
    child_stdin.write(_getdecoder().rgbout())
    child_stdin.close()
    return thumbimgpath

def makethumbsforclip(df):
    """takes a desc file and saves off thumbnail versions of all
    frames that have marks. the thumbnails go in a directory called
    thumbs/ off the same directory as the clip file. thumbnails that
    already exist are skipped. (delete them yourself if you want them
    to be regenerated). thumbnails are named
    </dir/to/clip>/thumbs/<fullclipfilename>.framenum.jpg"""

    clippath,clipname=os.path.split(df.clip().essencefile().filename())

    for m in df.clip().marks().getall():
        frame=m.getposition()[0]
        makethumbforframe(df,frame)

