import xml.dom.minidom as minidom
from xml.dom.ext import PrettyPrint
from xml.dom.ext.Visitor import Visitor, PreOrderWalker
import types,new
from cuisine.utils import *
import logging
log=logging.getLogger("Base")

"""
todo:

the xmlnodeclass xml helper functions are all over the place. i need
to systematically figure out what methods are needed and work out
proper names for them, then write them in an organized way.

"""


class xmlnodeclass(minidom.Element):
    """this is the virtual base class for classes that will correspond
    to xml nodes. The class name will be the same as the xml tag name,
    except the python class will have an uppercase first letter. most
    methods are underscored so as not to disturb your subclasses so
    much (and because they're for internal use). not all methods are
    underscored."""
    def __init__(self,alternateclassname=None):
        # the element tagname is the same as the subclass name, but
        # lowercased
        minidom.Element.__init__(self,alternateclassname or
                                 str(self.__class__.__name__).lower())

        # this is called when the node is new and when it's
        # "reanimated" from xml
        self.readxml()
 
    def _typenode(self,elem,nodetype):
        """if the given minidom is not of the python class type that
        we expected, we'll fix it. the subclassed element is returned."""
        if not isinstance(elem,nodetype):
            log.debug("recasting %s to %s" % (elem.__class__.__name__,
                                              nodetype.__name__))

            # should warn about funny casts (that aren't from
            # minidom.Element) here
            elem.__class__=nodetype
            elem.readxml()
        return elem

    def _synctree(self):
        """walks the tree, calling syncxml on all nodes that have it"""
        class SyncVisitor(Visitor):
            def visit(self, node):
                if hasattr(node,'syncxml'):
                    node.syncxml()
        
        PreOrderWalker(SyncVisitor(),self).run()
 
    def getChildrenElementsByTagName(self,name):
        """minidom forgot this one, apparently. getElementsByTagName
        does a recursive search"""
        ret=[]
        for node in self.childNodes:
            if (node.nodeType==minidom.Node.ELEMENT_NODE and
                (name=='*' or node.tagName==name)):
                ret.append(node)
        return ret

    def _getsetorcreate1(self,nodetype,setnode=None):
        """this returns our child xmlnodeclass with the name
        'nodetype', creating a new child if we don't already have one
        (that's what the 1 in the name is for. there will always be no
        more than one of any child element name/type in the
        document. if you pass in setnode, that node becomes the one
        element; any existing element is tossed.

        if nodetype is a string, the element will be fetched/created
        as usual, but it will be cast to a raw xmlnodeclass. this is
        useful for elements that don't deserve a python class, but you
        still want to be able to use the xmlnodeclass helpers on them
        (and within them).

        versions will probably happen here- when we request an
        element, it might be within a version tag. somehow, the app
        will have to indicate which version it wants to access or
        create. remember that versions are optional.
        """
        ret=self._getorset1(nodetype,setnode)
        if ret is None:
            # create
            if isinstance(nodetype,types.StringTypes):
                # no special class exists- use the vanilla type
                elem=xmlnodeclass(alternateclassname=nodetype) 
            else:
                elem=nodetype()
            log.debug("created new %s node" % (elem.__class__.__name__))
            self.appendChild(elem)
            
            # this would be minidom's job, but this elem isn't any
            # kind of Element subclass that minidom knows about
            elem.ownerDocument = self.ownerDocument
            
            ret=elem
        
        return ret

    def _getorset1(self,nodetype,setnode=None):
        """this is like _getsetorcreate1, but it won't create a new
        node. if there's no matching node, this version returns
        None. to create the node, you'd have to instance the new node
        yourself and then call this function again with the new node.
        """
        if isinstance(nodetype,str):
            # see method docstring for this case
            tagname=nodetype
            nodetype=xmlnodeclass
        else:
            # usually, we compute the xml tag name based on the class name
            tagname=str(nodetype.__name__).lower()
        elem=self.getChildrenElementsByTagName(tagname)
        if len(elem)>1:
            raise ValueError("surprised to find %i elements; expected 0 or 1" %
                             len(elem))

        if setnode is not None:
            # set
            if len(elem)==1:
                self.removeChild(elem[1])
            setnode.ownerDocument=self.ownerDocument
            self.appendChild(setnode)
            return setnode
        
        if len(elem)==0:
            # node not found
            return None
        else:
            # get
            elem=elem[0]
            # promote to a nodetype object if needed
            elem=self._typenode(elem,nodetype) 

        return elem
        
    
    def _getorsetattr(self,attrname,value=None):
        """get/set an attribute for this element tag. i think minidom
        returns '' if no attr found."""
        if value is None:
            return self.getAttribute(attrname)
        else:
            self.setAttribute(attrname,str(value))
            return self.getAttribute(attrname)

    def _getorsetstringnode(self,nodename,newval=None,**attrs):
        
        """for a string subnode that is not yet a class of its own, we
        have this function to get/set the text in the node. doesn't
        support multiply-listed child nodes. the _string_ is returned
        in any case.

        this will make or update a single child node like
        <nodename>newval</nodename>.

        if you're setting the value, you can also throw in some
        attributes as keyword args.  """
        
        nodes=self.getChildrenElementsByTagName(nodename)

        if newval is None:
            # get
            if len(nodes)==0:
                return None
            return self._getorsettextfornode(nodes[0])

        # set
        if len(nodes)==0:
            # create the new node
            node=minidom.Element(nodename)
            node.ownerDocument=self.ownerDocument
            self.appendChild(node)
        else:
            node=nodes[0]

        # set any given node attributes (on top of whatever may have
        # been there)
        
        for k,v in attrs.items():
            self.setAttribute(str(k),str(v))
        # set the text
        return self._getorsettextfornode(node,newval)

    def _getorsettextfornode(self,node,newval=None):
        
        """given an Element node, this gets the text within (like in
        <node>text</node>) or sets it to a new value. new values are
        str'd, but no decoding is performed.

        note that we're not checking the contents of the node-- the
        data of the first child of the node is what's getting
        returned. a set involves removing the first child node (if
        there is one) and then inserting a text node.

        not sure why you have to pass the node in (instead of reading
        self), but maybe the code uses that feature somewhere.  """
        
        if newval is None:
            # get
            return node.firstChild.data
        else:
            # set
            
            # (removes the first child node (if present) and addes a
            # text node-- don't call this on the wrong element!)

            if node.hasChildNodes():
                node.removeChild(node.firstChild)
            txt=minidom.Text()
            txt.data=str(newval)
            node.appendChild(txt)
            return txt.data
        
  
    def _getorsettypednode(self,nodename,nodetype,newval=None,**attrs):
        """
        this is like _getorsetstringnode, but it serializes data to
        and from the type 'nodetype' (which is probably in
        desctypes.py). I.e. newval must be of type 'nodetype',
        returned values will always be of type 'nodetype'. nodetype is
        expected to have str/__init__ methods that serialize to a
        string and recreate the value """

        if newval is None:
            # get
            val=self._getorsetstringnode(nodename)
            if val is None:
                return None
            else:
                return nodetype(val)
        else:
            # set
            if isinstance(newval,nodetype):
                self._getorsetstringnode(nodename,str(newval),**attrs)
            else:
                raise TypeError("'%r' is not a %s, won't set node value" %
                                (newval,nodetype.__name__))
            return newval
                           
    def _getorsettypedattr(self,attrname,attrtype,newval=None):
        """
        gets/sets an attribute for this node, converting to/from the
        given type just like _getorsettypednode
        """
        if newval is None:
            # get
            return attrtype(self._getorsetattr(attrname))
        else:
            # set
            if isinstance(newval,attrtype):
                self._getorsetattr(attrname,str(newval))
            else:
                raise TypeError("'%r' is not a %s, won't set node value" %
                                (newval,nodetype.__name__))
            return newval
        
    def uuid(self):
        """all nodes, for now, have a uuid read-only accessor"""
        return self.getAttribute("uuid")
    def assignuuid(self):
        
        """gives this element a uuid='...' attribute and a read-only
        accessor"""

        self.setAttribute("uuid",createuuid())
#        def _(self): # read-only
#            return self.getAttribute('uuid')
#        self.uuid=new.instancemethod(_,self,self.__class__)
#        print "brick uuid",self.uuid()

    def parentbytype(self,nodetype):
        """this returns a parent node by walking up the tree
        and returning the first node that has the type given by
        (or element named by) nodetype.

        parentbytype(Clip) or parentbytype('clip') (note lower case)
        are almost the same.  they might be different if there was no
        class called Clip so the clip node stayed as a
        minidom.Element. So the latter form, although more error-prone
        since you could misspell the string, is probably more robust.

        returns None if no matching parent is found
        """
        n=self
        if not isinstance(nodetype,(str,unicode)):
            # match by type
            test=lambda n: isinstance(n,nodetype) 
        else:
            # match by name
            test=lambda n: n.__class__.__name__.lower()==nodetype 
        while n and not test(n):
            n=n.parentNode
        return n

    def readxml(self):
        """this gets called when the xml has been loaded from
        somewhere and the object might want to create an internal
        non-xml representation for performance reasons. readxml() is
        called after __init__; you don't need to do it
        yourself. syncxml() will be called when the internal
        representation needs to be converted back into xml.
        you're on your own to maintian the sync. 
        this
        method is entirely optional."""
        pass
    def syncxml(self):
        """if you keep internal versions of your data and avoid
        writing them to xml all the time (see Channel for an example),
        update the xml in this method.  this gets called before saves
        and other times when we come to look at your xml. this method
        is entirely optional."""
        pass

    def copy(self):
        """makes a copy of this node which shall be a literal copy of all
        the xml within it.

        since you're calling copy(), we know there was a python class
        for this node. the returned version will also be turned into a
        python class, although its children won't be (but they will
        convert when you visit them).

        even uuids will be duplicated, which might not be
        desirable. it probably is copy()'s job to do some new uuids,
        but i haven't implemented that yet.

        this probably shouldn't be overridden too much, except maybe
        to handle the uuids.

        the copy isn't inserted in the tree anywhere (to my knowledge)
        """

        self._synctree()

        ##         class SyncVisitor(Visitor):
        ##             def visit(self, node):
        ##                 print node,node.ownerDocument

        ##         PreOrderWalker(SyncVisitor(),self).run()


        #        PrettyPrint(self)

        # oh how scary--

        selfxml=self.toxml()
        newnode=minidom.parseString(selfxml).documentElement.cloneNode(deep=1)

        # this will call readxml if necc
        newnode=self._typenode(newnode,self.__class__) 

        return newnode

class collectiveelement(xmlnodeclass):
    """specialized element that just holds a bunch of one other element
    (maybe this should define an 'x in y' method?)
    """

    def childtype(self):

        """this is a constant for each subclass. the subclass just
        needs to return the type for all of the collective's
        children"""
        
        raise NotImplementedError("subclass needs to return the type for "
                                  "its children")

    def getall(self):  
        """if all your child elements are the same type, you can use
        this to fetch them.  we could almost figure out the child node
        type by capitalizing the element tag, but you'd have to pass
        your globals() in to make the right types be available, i
        think. coming in version 2."""
        # we skip non-Elements, which eliminates the text nodes that appear
        # after a prettyprint
        return [self._typenode(x,self.childtype()) for x in self.childNodes
                if isinstance(x,minidom.Element)]

    def add(self,newelem):
        """adds a new element to the child list, forcing it to the our
        child type if necessary"""
        if newelem.ownerDocument not in (self.ownerDocument,None):
            log.error("%r was already owned by %r - taking ownership" %
                      (newelem,newelem.ownerDocument))
        ch=self._typenode(newelem,self.childtype())
        ch.ownerDocument=self.ownerDocument
        self.appendChild(ch)
        return newelem
    def new(self):
        """like add, but creates the new element too (and returns it)"""
        elem=(self.childtype())()
        self.add(elem)
        return elem
    def remove(self,child):
        matches=[x for x in self.childNodes if x==child]
        if len(matches)==0:
            raise ValueError("child %r not found in this collection" % child)
        self.removeChild(child)

class dictelement(xmlnodeclass):
    """specialized element for blocks like this:

    <foos>
      <foo key='key' value='value'/>
    </foos>

    you probably didn't make classes for Foos and Foo. the dict
    element can be accessed just like a dict. the key and value are
    treated as unit'd types, so they go through str() and
    decodeunit().  """
    
    def elementname(self):
        raise NotImplementedError("subclass must return the element name for "
                                  "the child tags")
    
        # (we're just going to put an 's' on the child tag name to
        # make the parent tag name)
    

class xmldocfile:
    """this is the base class for our documents (descfiles, showfiles)"""
    def _openornew(self,filename,topleveltype=None):
        """opens the file into self._doc or, if the file doesn't
        exist, sets self._doc to an empty minidom.Document with a new
        topleveltype object (such as Clip or Show)"""

        self.docfilename=filename
        if os.path.exists(self.docfilename):
            f=open(self.docfilename)
            log.debug("reading document from existing file '%s'",
                      self.docfilename)
            self._doc=None
            try:
                self._doc=minidom.parseString(f.read())
            except xml.parsers.expat.ExpatError,e:
                raise IOError("error reading XML file '%s': %s"%
                              (self.docfilename,e.args[0]))
            f.close()
        else: # doesn't exist
            log.info("starting new document (%s)"%self.docfilename)
            self._doc=minidom.Document()
            toplevelnode=topleveltype()
            toplevelnode.ownerDocument = self._doc
            self._doc.appendChild(toplevelnode)

        topNodeName = self._gettoplevel().localName
        if topNodeName != topleveltype.__name__.lower():
                raise ValueError("XML outmost node is type %s; cannot make a %s object from it" % (topNodeName, topleveltype.__name__))

        # jam the top level to the proper type
        self._doc.documentElement.__class__=topleveltype

    def _gettoplevel(self):
        """returns the document's main node"""
        return self._doc.documentElement

    def _sync(self):
        """call the syncxml method on each xmlnodeclass object, so it
        can sync any internal data to the xml. for performance
        reasons, some nodes might keep their own data structures and
        only update their xml attributes and children on demand. (this
        is another reason you shouldn't ask xmlnodeclass objects for
        their children directly-- they might be out of date"""

        # i hope at least the toplevel is an xmlnodeclass..
        self._gettoplevel()._synctree() 
        
    def pprint(self):
        self._sync()
        PrettyPrint(self._doc)
        
    def save(self):

        """writes this document back to disk in the same filename it
        was loaded from. this should use some form of safe saving, to
        avoid ever writing zero-length files."""

        log.info("saving %s to %s" %
                 (self._gettoplevel().__class__.__name__, self.docfilename))
        self._sync()
        f=open(self.docfilename,"wb")
        PrettyPrint(self._doc,stream=f)
#        f.write(self._doc.toprettyxml())
        f.close()
    def getfilename(self):
        return self.docfilename

try:
    import psyco
    
    # note that there's class-changing all over, which psyco claims
    # not to like.
    
    psyco.bind(xmlnodeclass)
    psyco.bind(collectiveelement)
    psyco.bind(xmldocfile)
except ImportError:
    pass
